import hashlib

# import json
from typing import Any, Dict, List, Union

from .constants import (  # TYPE_PARAMETER_OPTIONS_DICT,
    ESTIMATE,
    TYPE_PARAMETER_DICT,
    TYPE_PARAMETER_OPTIONS,
)


def hashify(x: Any) -> str:
    """Generate a hash.

    Args:
        x:
            The object to hash. `x` will be converted to a string (`str(x)`), then
            hashed.

    Returns:
        The hash.
    """
    # return int(hashlib.sha256(str(x).encode('utf-8')).hexdigest(), 16)
    return hashlib.blake2b(str(x).encode('utf-8')).hexdigest()


def hash_parameter_dict(dict_: TYPE_PARAMETER_DICT):
    """Hash a dictionary of parameter values."""
    value = tuple(zip(dict_.keys(), dict_.values()))
    return hashify(value)


def hash_parameter_options(list_: TYPE_PARAMETER_OPTIONS):
    """Hash parameter options."""
    return hashify(list(list_))


def hash_str(str_: str):
    return hashify(str_)


def hash_list(list_: List):
    return hashify(list(list_))


def snake_case_to_camel_case(string: str) -> str:
    """Convert a string from snake case to camel case.

    Args:
        string:
            The string, in snake case.

    Returns:
        The string, in camel case.
    """
    string_pieces = string.split('_')
    string_camel = ''
    for string_piece in string_pieces:
        string_camel += string_piece[0].upper() + string_piece[1:]
    return string_camel


def parameter_string_to_value(
    parameter_string: str,
    passthrough_estimate: bool = False,
) -> Union[float, int, str]:
    """Cast a parameter value from string to numeric.

    Args:
        parameter_string:
            The parameter value, as a string.
        passthrough_estimate:
            Whether to return `ESTIMATE` as `ESTIMATE`. If `False`, raises an exception
            if `parameter_string == ESTIMATE`.

    Returns:
        The parameter value, as a numeric type.
    """
    if parameter_string == ESTIMATE:
        if passthrough_estimate:
            return parameter_string
        raise ValueError('Please handle estimated parameters differently.')
    float_value = float(parameter_string)
    int_value = int(float_value)

    if int_value == float_value:
        return int_value
    return float_value


# def hash_dictionary(dictionary: Dict[str, Union[]]):
#    return hash(json.dumps(dictionary, sort_keys=True))
#
#
# def hash_list(list_: List):
#    return hash(json.dumps(list_, sort_keys=True))



"""The model selection problem class."""
import abc
from functools import partial
from itertools import chain
from pathlib import Path
from typing import Any, Callable, Dict, Iterable, Optional, Union

import yaml

from .candidate_space import CandidateSpace, method_to_candidate_space_class
from .constants import (
    CANDIDATE_SPACE_ARGUMENTS,
    CRITERION,
    METHOD,
    MODEL_SPACE_FILES,
    PREDECESSOR_MODEL,
    VERSION,
    Criterion,
    Method,
)
from .model import Model, default_compare
from .model_space import ModelSpace


class Problem(abc.ABC):
    """Handle everything related to the model selection problem.

    Attributes:
        model_space:
            The model space.
        calibrated_models:
            Calibrated models. Will be used to augment the model selection problem (e.g.
            by excluding them from the model space). FIXME refactor out
        candidate_space_arguments:
            Custom options that are used to construct the candidate space.
        compare:
            A method that compares models by selection criterion. See
            `petab_select.model.default_compare` for an example.
        criterion:
            The criterion used to compare models.
        method:
            The method used to search the model space.
        version:
            The version of the PEtab Select format.
        yaml_path:
            The location of the selection problem YAML file. Used for relative
            paths that exist in e.g. the model space files.
            TODO should the relative paths be relative to the YAML or the
            file that contains them?

    Unsaved attributes:
        candidate_space:
            The candidate space that will be used.
            Reason for not saving:
                Essentially reproducible from `Problem.method` and
                `Problem.calibrated_models`.
    """

    def __init__(
        self,
        model_space: ModelSpace,
        candidate_space_arguments: Dict[str, Any] = None,
        compare: Callable[[Model, Model], bool] = None,
        criterion: Criterion = None,
        method: str = None,
        version: str = None,
        yaml_path: str = None,
    ):
        self.model_space = model_space
        self.criterion = criterion
        self.method = method
        self.version = version
        self.yaml_path = yaml_path

        self.candidate_space_arguments = candidate_space_arguments
        if self.candidate_space_arguments is None:
            self.candidate_space_arguments = {}

        self.compare = compare
        if self.compare is None:
            self.compare = partial(default_compare, criterion=self.criterion)

    def get_path(self, relative_path: Union[str, Path]) -> Path:
        """Get the path to a resource, from a relative path.

        Args:
            relative_path:
                The path to the resource, that is relative to the PEtab Select
                problem YAML file location.

        Returns:
            The path to the resource.

        TODO:
            Unused?
        """
        if self.yaml_path is None:
            return Path(relative_path)
        return self.yaml_path.parent / relative_path

    def exclude_models(
        self,
        models: Iterable[Model],
    ) -> None:
        """Exclude models from the model space.

        Args:
            models:
                The models.
        """
        self.model_space.exclude_models(models)

    def exclude_model_hashes(
        self,
        model_hashes: Iterable[str],
    ) -> None:
        """Exclude models from the model space, by model hashes.

        Args:
            models:
                The model hashes.
        """
        self.model_space.exclude_model_hashes(model_hashes)

    @staticmethod
    def from_yaml(
        yaml_path: Union[str, Path],
    ) -> 'Problem':
        """Generate a problem from a PEtab Select problem YAML file.

        Args:
            yaml_path:
                The location of the PEtab Select problem YAML file.

        Returns:
            A `Problem` instance.
        """
        yaml_path = Path(yaml_path)
        with open(yaml_path, 'r') as f:
            problem_specification = yaml.safe_load(f)

        if not problem_specification.get(MODEL_SPACE_FILES, []):
            raise KeyError(
                'The model selection problem specification file is missing '
                'model space files.'
            )

        model_space = ModelSpace.from_files(
            # problem_specification[MODEL_SPACE_FILES],
            [
                # `pathlib.Path` appears to handle absolute `model_space_file` paths
                # correctly, even if used as a relative path.
                # TODO test
                # This is similar to the `Problem.get_path` method.
                yaml_path.parent / model_space_file
                for model_space_file in problem_specification[
                    MODEL_SPACE_FILES
                ]
            ],
            # source_path=yaml_path.parent,
        )

        criterion = problem_specification.get(CRITERION, None)
        if criterion is not None:
            criterion = Criterion(criterion)

        candidate_space_arguments = problem_specification.get(
            CANDIDATE_SPACE_ARGUMENTS,
            None,
        )
        if candidate_space_arguments is not None:
            if PREDECESSOR_MODEL in candidate_space_arguments:
                candidate_space_arguments[PREDECESSOR_MODEL] = (
                    yaml_path.parent
                    / candidate_space_arguments[PREDECESSOR_MODEL]
                )

        return Problem(
            model_space=model_space,
            candidate_space_arguments=candidate_space_arguments,
            criterion=criterion,
            # TODO refactor method to use enum
            method=problem_specification.get(METHOD, None),
            version=problem_specification.get(VERSION, None),
            yaml_path=yaml_path,
        )

    def get_best(
        self,
        models: Optional[Iterable[Model]],
        criterion: Optional[Union[str, None]] = None,
        compute_criterion: bool = False,
    ) -> Model:
        """Get the best model from a collection of models.

        The best model is selected based on the selection problem's criterion.

        Args:
            models:
                The best model will be taken from these models.
            criterion:
                The criterion by which models will be compared. Defaults to
                `self.criterion` (e.g. as defined in the PEtab Select problem YAML
                file).
            compute_criterion:
                Whether to try computing criterion values, if sufficient
                information is available (e.g., likelihood and number of
                parameters, to compute AIC).

        Returns:
            The best model.
        """
        if criterion is None:
            criterion = self.criterion

        best_model = None
        for model in models:
            if compute_criterion and not model.has_criterion(criterion):
                model.get_criterion(criterion)
            if best_model is None:
                if model.has_criterion(criterion):
                    best_model = model
                # TODO warn if criterion is not available?
                continue
            if self.compare(best_model, model, criterion=criterion):
                best_model = model
        if best_model is None:
            raise KeyError(
                f'None of the supplied models have a value set for the criterion {criterion}.'
            )
        return best_model

    def new_candidate_space(
        self,
        *args,
        method: Method = None,
        **kwargs,
    ) -> None:
        """Construct a new candidate space.

        Args:
            *args, **kwargs:
                Arguments are passed to the candidate space constructor.
            method:
                The model selection method.
        """
        if method is None:
            method = self.method
        candidate_space_class = method_to_candidate_space_class(method)
        candidate_space_arguments = (
            candidate_space_class.read_arguments_from_yaml_dict(
                self.candidate_space_arguments
            )
        )
        candidate_space_kwargs = {
            **candidate_space_arguments,
            **kwargs,
        }
        candidate_space = candidate_space_class(
            *args,
            **candidate_space_kwargs,
        )
        return candidate_space



"""The PEtab Select command-line interface."""
import warnings
from pathlib import Path
from typing import Any, Dict, List

import click
import dill
import numpy as np
import pandas as pd
import yaml
from more_itertools import one

from . import ui
from .candidate_space import CandidateSpace, method_to_candidate_space_class
from .constants import INITIAL_MODEL_METHODS, PETAB_YAML
from .model import Model, models_from_yaml_list, models_to_yaml_list
from .problem import Problem


def read_state(filename: str) -> Dict[str, Any]:
    with open(filename, 'rb') as f:
        state = dill.load(f)

    state['problem'] = dill.loads(state['problem'])
    state['candidate_space'] = dill.loads(state['candidate_space'])

    return state


def write_state(
    state: Dict[str, Any],
    filename: str,
) -> Dict[str, Any]:
    with open(filename, 'wb') as f:
        dill.dump(state, f)


def get_state(
    problem: Problem,
    candidate_space: CandidateSpace,
) -> Dict[str, Any]:
    state = {
        'problem': dill.dumps(problem),
        'candidate_space': dill.dumps(candidate_space),
    }
    return state


def check_state_compatibility(
    state: Dict[str, Any],
    problem: Problem,
    candidate_space: CandidateSpace,
):
    # if state['problem'].method != problem.method:
    #    warnings.warn(
    #        'The method in the problem loaded from the state does not match '
    #        'the method specified in either the PEtab Select YAML file or '
    #        'specified explicitly with this command.'
    #    )
    pass


@click.group()
def cli():
    pass


@cli.command("candidates")
@click.option(
    '--problem',
    '-p',
    'problem_yaml',
    help='The PEtab Select YAML problem file.',
)
@click.option(
    '--state',
    '-s',
    'state_filename',
    type=str,
    help='The file that stores the state.',
)
@click.option(
    '--output',
    '-o',
    'output',
    type=str,
    help='The file where candidate models will be stored.',
)
@click.option(
    '--method',
    '-m',
    'method',
    type=str,
    default=None,
    help='The method used to identify the candidate models. Defaults to the method in the problem YAML.',
)
@click.option(
    '--previous-predecessor-model',
    '-P',
    'previous_predecessor_model_yaml',
    type=str,
    default=None,
    help='(Optional) The predecessor model used in the previous iteration of model selection.',
)
@click.option(
    '--calibrated-models',
    '-C',
    'calibrated_models_yamls',
    type=str,
    multiple=True,
    default=None,
    help='(Optional) Models that have been calibrated.',
)
@click.option(
    '--newly-calibrated-models',
    '-N',
    'newly_calibrated_models_yamls',
    type=str,
    multiple=True,
    default=None,
    help=(
        '(Optional) Models that were calibrated in the most recent iteration.'
    ),
)
@click.option(
    '--limit',
    '-l',
    'limit',
    type=float,
    default=np.inf,
    help='(Optional) Limit the number of models in the output.',
)
@click.option(
    '--limit-sent',
    '-L',
    'limit_sent',
    type=float,
    default=np.inf,
    help=(
        '(Optional) Limit the number of models sent to the candidate space '
        '(which are possibly rejected and excluded from the output).'
    ),
)
@click.option(
    '--relative-paths/--absolute-paths',
    'relative_paths',
    type=bool,
    default=False,
    help='Whether to output paths relative to the output file.',
)
@click.option(
    '--excluded-models',
    '-e',
    'excluded_model_files',
    type=str,
    multiple=True,
    default=None,
    help='Exclude models in this file.',
)
@click.option(
    '--excluded-model-hashes',
    '-E',
    'excluded_model_hash_files',
    type=str,
    multiple=True,
    default=None,
    help='Exclude model hashes in this file (one model hash per line).',
)
def candidates(
    problem_yaml: str,
    state_filename: str,
    output: str,
    method: str = None,
    previous_predecessor_model_yaml: str = None,
    # best: str = None,
    calibrated_models_yamls: List[str] = None,
    newly_calibrated_models_yamls: List[str] = None,
    limit: float = np.inf,
    limit_sent: float = np.inf,
    relative_paths: bool = False,
    excluded_model_files: List[str] = None,
    excluded_model_hash_files: List[str] = None,
) -> None:
    """Search for candidate models in the model space.

    Documentation for arguments can be viewed with
    `petab_select candidates --help`.
    """
    problem = Problem.from_yaml(problem_yaml)
    if method is None:
        method = problem.method

    # `petab_select.ui.candidates` uses `petab_select.Problem.method` to generate
    # the candidate space.
    problem.method = method

    candidate_space = problem.new_candidate_space(limit=limit)

    # Setup state
    if not Path(state_filename).exists():
        Path(state_filename).parent.mkdir(parents=True, exist_ok=True)
    else:
        state = read_state(state_filename)
        check_state_compatibility(
            state=state,
            problem=problem,
            candidate_space=candidate_space,
        )
        problem = state['problem']
        candidate_space = state['candidate_space']

    excluded_models = []
    # TODO seems like default is `()`, not `None`...
    if excluded_model_files is not None:
        for model_yaml_list in excluded_model_files:
            excluded_models.extend(models_from_yaml_list(model_yaml_list))

    # TODO test
    excluded_model_hashes = []
    if excluded_model_hash_files is not None:
        for excluded_model_hash_file in excluded_model_hash_files:
            with open(excluded_model_hash_file, 'r') as f:
                excluded_model_hashes += f.read().split('\n')

    previous_predecessor_model = candidate_space.predecessor_model
    if previous_predecessor_model_yaml is not None:
        previous_predecessor_model = Model.from_yaml(
            previous_predecessor_model_yaml
        )

    # FIXME write single methods to take all models from lists of lists of
    #       models recursively
    calibrated_models = None
    if calibrated_models_yamls is not None:
        calibrated_models = {}
        for calibrated_models_yaml in calibrated_models_yamls:
            calibrated_models.update(
                {
                    model.get_hash(): model
                    for model in models_from_yaml_list(calibrated_models_yaml)
                }
            )

    newly_calibrated_models = None
    if newly_calibrated_models_yamls is not None:
        newly_calibrated_models = {}
        for newly_calibrated_models_yaml in newly_calibrated_models_yamls:
            newly_calibrated_models.update(
                {
                    model.get_hash(): model
                    for model in models_from_yaml_list(
                        newly_calibrated_models_yaml
                    )
                }
            )

    ui.candidates(
        problem=problem,
        candidate_space=candidate_space,
        previous_predecessor_model=previous_predecessor_model,
        calibrated_models=calibrated_models,
        newly_calibrated_models=newly_calibrated_models,
        limit=limit,
        limit_sent=limit_sent,
        excluded_models=excluded_models,
        excluded_model_hashes=excluded_model_hashes,
    )

    # Save state
    write_state(
        state=get_state(
            problem=problem,
            candidate_space=candidate_space,
        ),
        filename=state_filename,
    )

    # Save candidates
    models_to_yaml_list(
        models=candidate_space.models,
        output_yaml=output,
        relative_paths=relative_paths,
    )


@cli.command("model_to_petab")
@click.option(
    '--model',
    '-m',
    'models_yamls',
    multiple=True,
    help='The PEtab Select model YAML file.',
)
@click.option(
    '--output',
    '-o',
    'output_path',
    type=str,
    help='The directory where the PEtab files will be output.',
)
@click.option(
    '--model_id',
    '-i',
    'model_id',
    type=str,
    default=None,
    help=(
        '(Optional) The ID of the model to use, in case the YAML file '
        'contains multiple models.'
    ),
)
def model_to_petab(
    models_yamls: List[str],
    output_path: str,
    model_id: str = None,
) -> None:
    """Create a PEtab problem from a PEtab Select model YAML file.

    The filename for the PEtab problem YAML file is output to `stdout`.

    Documentation for arguments can be viewed with
    `petab_select model_to_petab --help`.
    """
    models = []
    for models_yaml in models_yamls:
        models.extend(models_from_yaml_list(models_yaml))

    model0 = None
    try:
        model0 = one(models)
    except:
        for model in models:
            if model.model_id == model_id:
                if model0 is not None:
                    raise ValueError('There are multiple models with this ID.')
                model0 = model
                # TODO could `break` here and remove the above `ValueError`
                #      and the `model0` logic
    if model0 is None:
        raise ValueError('Could not find a model with the specified model ID.')

    if model_id is not None and model0.model_id != model_id:
        raise ValueError(
            'The ID of the model from the YAML file does not match the '
            'specified ID.'
        )

    result = ui.model_to_petab(model0, output_path)
    print(result[PETAB_YAML])


@cli.command("models_to_petab")
@click.option(
    '--models',
    '-m',
    'models_yamls',
    type=str,
    multiple=True,
    help='The PEtab Select model YAML file, containing a list of models.',
)
@click.option(
    '--output',
    '-o',
    'output_path_prefix',
    type=str,
    help='The directory where the PEtab files will be output. The PEtab files will be stored in a model-specific subdirectory.',
)
def models_to_petab(
    models_yamls: List[str],
    output_path_prefix: str,
) -> None:
    """Create a PEtab problem for each model in a PEtab Select model YAML file.

    NB: Models in the YAML file must have a model ID.

    The output to `stdout` is a two-column tab-separated list, where the first
    column is the model ID, and the second column is the location of the PEtab
    problem YAML file for that model.

    Documentation for arguments can be viewed with
    `petab_select models_to_petab --help`.
    """
    models = []
    for models_yaml in models_yamls:
        models.extend(models_from_yaml_list(models_yaml))

    model_ids = pd.Series([model.model_id for model in models])
    duplicates = '\n'.join(set(model_ids[model_ids.duplicated()]))
    if duplicates:
        raise ValueError(
            'It appears that the provided PEtab Select model YAML file '
            'contains multiple models with the same ID. The following '
            f'duplicates were detected: {duplicates}'
        )

    results = ui.models_to_petab(
        models,
        output_path_prefix=output_path_prefix,
    )
    result_string = '\n'.join(
        [
            '\t'.join([model.model_id, result[PETAB_YAML]])
            for model, result in zip(models, results)
        ]
    )
    print(result_string)


@cli.command("best")
@click.option(
    '--problem',
    '-p',
    'problem_yaml',
    type=str,
    help='The PEtab Select YAML problem file.',
)
@click.option(
    '--models',
    '-m',
    'models_yamls',
    type=str,
    multiple=True,
    help='A list of calibrated models.',
)
@click.option(
    '--output',
    '-o',
    'output',
    type=str,
    help='The file where the best model will be stored.',
)
@click.option(
    '--state',
    '-s',
    'state_filename',
    type=str,
    default=None,
    help='The file that stores the state.',
)
@click.option(
    '--criterion',
    '-c',
    'criterion',
    type=str,
    default=None,
    help='The criterion by which models will be compared.',
)
@click.option(
    '--relative-paths/--absolute-paths',
    'relative_paths',
    type=bool,
    default=False,
    help='Whether to output paths relative to the output file.',
)
def best(
    problem_yaml: str,
    models_yamls: List[str],
    output: str,
    state_filename: str = None,
    criterion: str = None,
    relative_paths: bool = False,
) -> None:
    """Get the best model from a list of models.

    Documentation for arguments can be viewed with
    `petab_select get_best --help`.
    """
    paths_relative_to = None
    if relative_paths:
        paths_relative_to = Path(output).parent

    problem = Problem.from_yaml(problem_yaml)

    if state_filename is not None:
        state = read_state(state_filename)
        check_state_compatibility(
            state=state,
            problem=problem,
            candidate_space=None,
        )
        set_state(
            state=state,
            problem=problem,
            candidate_space=None,
        )

    models = []
    for models_yaml in models_yamls:
        models.extend(models_from_yaml_list(models_yaml))

    best_model = ui.best(
        problem=problem,
        models=models,
        criterion=criterion,
    )
    best_model.to_yaml(output, paths_relative_to=paths_relative_to)


cli.add_command(candidates)
cli.add_command(model_to_petab)
cli.add_command(models_to_petab)
cli.add_command(best)



from pathlib import Path
from typing import List, Optional

import petab
from petab.C import ESTIMATE, NOMINAL_VALUE

from .constants import PETAB_ESTIMATE_FALSE, TYPE_PARAMETER_DICT, TYPE_PATH


class PetabMixin:
    """Useful things for classes that contain a PEtab problem.

    All attributes/methods are prefixed with `petab_`.
    """

    def __init__(
        self,
        petab_yaml: Optional[TYPE_PATH] = None,
        petab_problem: Optional[petab.Problem] = None,
        parameters_as_lists: bool = False,
    ):
        if petab_yaml is None and petab_problem is None:
            raise ValueError(
                'Please supply at least one of either the location of the '
                'PEtab problem YAML file, or an instance of the PEtab problem.'
            )
        self.petab_yaml = petab_yaml
        if self.petab_yaml is not None:
            self.petab_yaml = Path(self.petab_yaml)

        self.petab_problem = petab_problem
        if self.petab_problem is None:
            self.petab_problem = petab.Problem.from_yaml(str(petab_yaml))

        self.petab_parameters = {
            parameter_id: (
                row[NOMINAL_VALUE]
                if row[ESTIMATE] == PETAB_ESTIMATE_FALSE
                else ESTIMATE
            )
            for parameter_id, row in self.petab_problem.parameter_df.iterrows()
        }
        if parameters_as_lists:
            self.petab_parameters = {
                k: [v] for k, v in self.petab_parameters.items()
            }

    @property
    def petab_parameter_ids_estimated(self) -> List[str]:
        return [
            parameter_id
            for parameter_id, parameter_value in self.petab_parameters.items()
            if parameter_value == ESTIMATE
        ]

    @property
    def petab_parameter_ids_fixed(self) -> List[str]:
        estimated = self.petab_parameter_ids_estimated
        return [
            parameter_id
            for parameter_id in self.petab_parameters
            if parameter_id not in estimated
        ]

    @property
    def petab_parameters_singular(self) -> TYPE_PARAMETER_DICT:
        return {
            parameter_id: one(parameter_value)
            for parameter_id, parameter_value in self.petab_parameters
        }
